
#define TBLOCK 512
#define NAMSIZ 100
union hblock {
  char dummy[TBLOCK];
  struct header {
    char name[NAMSIZ];
    char mode[8];
    char uid[8];
    char gid[8];
    char size[12];
    char mtime[12];
    char chksum[8];
    char linkflag;
    char linkname[NAMSIZ];
  } dbuf;
};
